import os, shutil

FILE_FORMATS = {
    # Format.lower()    # File extension. By convention, use the first by default
    'hdf5':             ['hdf5', 'h5', 'hdf'],
    'tiff':             ['tiff', 'tif'],
    'png':              ['png'],
    'ramon':            ['m'],
    'matlab':           ['m', 'mat'],
}


# This is not simply FILE_FORMATS in 'reverse', because while we may be able
# to get the file extension for a RAMON object (.m) we can't guess a datatype
# from the ambiguous extension ".m".


def _fail_pair_conversion(i, o):
    """
    Helper-function to print failure and pass back False.
    """
    print("Conversion from {0} to {1} is not currently supported.".format(i, o))
    return False


def _get_extension_for_format(fmt):
    """
    Get the appropriate file extension for a given format.

    Arguments:
        fmt:        The format to find an extension for

    Returns:
        String. The format (without leading period),
                or False if none was found.
    """
    if fmt in FILE_FORMATS:
        # Use first file extension by default
        return FILE_FORMATS[fmt][0]

    # Otherwise, we don't recognize the format...
    return False



def _guess_format_from_extension(ext):
    """
    Guess the appropriate data type from file extension.

    Arguments:
        ext:        The file extension (period optional)

    Returns:
        String. The format (without leading period),
                or False if none was found or couldn't be guessed
    """
    ext = ext.strip('.')

    # We look through FILE_FORMATS for this extension.
    # - If it appears zero times, return False. We can't guess.
    # - If it appears once, we can simply return that format.
    # - If it appears more than once, we can't guess (it's ambiguous,
    #   e.g .m = RAMON or MATLAB)

    formats = []
    for fmt in FILE_FORMATS:
        if ext in FILE_FORMATS[fmt]:
            formats.append(fmt)

    if formats == [] or len(formats) > 1:
        return False

    return formats[0]



def convert(in_file, out_file, in_fmt="", out_fmt=""):
    """
    Converts in_file to out_file, guessing datatype in the absence of
    in_fmt and out_fmt.

    Arguments:
        in_file:    The name of the (existing) datafile to read
        out_file:   The name of the file to create with converted data
        in_fmt:     Optional. The format of incoming data, if not guessable
        out_fmt:    Optional. The format of outgoing data, if not guessable

    Returns:
        String. Output filename
    """

    # First verify that in_file exists and out_file doesn't.
    in_file = os.path.expanduser(in_file)
    out_file = os.path.expanduser(out_file)

    if os.path.exists(out_file):
        raise IOError("Output file {0} already exists, stopping...".format(out_file))
    if not os.path.exists(in_file):
        raise IOError("Input file {0} does not exist, stopping...".format(in_file))

    # Get formats, either by explicitly naming them or by guessing.
    # TODO: It'd be neat to check here if an explicit fmt matches the guess.
    in_fmt = in_fmt.lower() or _guess_format_from_extension(
             in_file.split('.')[-1].lower())
    out_fmt = out_fmt.lower() or _guess_format_from_extension(
              out_file.split('.')[-1].lower())

    if not in_fmt or not out_fmt:
        print("Cannot determine conversion formats.")
        return False

    if in_fmt is out_fmt:
        # This is the case when this module (intended for LONI) is used
        # indescriminately to 'funnel' data into one format.
        shutil.copyfileobj(in_file, out_file)
        return out_file


    ## Import
    if in_fmt == 'hdf5':
        import hdf5
        data = hdf5.import_hdf5(in_file)
    elif in_fmt == 'tiff':
        import tiff
        data = tiff.import_tiff(in_file)
    elif in_fmt == 'png':
        import png
        data = png.import_png(in_file)
    else:
        return _fail_pair_conversion(in_fmt, out_fmt)

    ## Export
    if out_fmt == 'hdf5':
        import hdf5
        return hdf5.export_hdf5(out_file, data)
    elif out_fmt == 'tiff':
        import tiff
        return tiff.export_tiff(out_file, data)
    elif out_fmt == 'png':
        import png
        return png.export_png(out_file, data)
    else:
        return _fail_pair_conversion(in_fmt, out_fmt)


    return _fail_pair_conversion(in_fmt, out_fmt)



import numpy


def snap_to_cube(q_start, q_stop, chunk_depth=16, q_index=1):
    """
    For any q in {x, y, z, t}
    Takes in a q-range and returns a 1D bound that starts at a cube
    boundary and ends at another cube boundary and includes the volume
    inside the bounds. For instance, snap_to_cube(2, 3) = (1, 17)

    Arguments:
        :q_start:       `int` The lower bound of the q bounding box of volume
        :q_stop:        `int` The upper bound of the q bounding box of volume
        :chunk_depth:   `int : CHUNK_DEPTH` The size of the chunk in this volume (use ``get_info()``)
        :q_index:       `int : 1` The starting index of the volume (in q)
    Returns:
        :2-tuple: ``(lo, hi)`` bounding box for the volume
    """
    lo = 0; hi = 0
    # Start by indexing everything at zero for our own sanity
    q_start -= q_index; q_stop -= q_index

    if q_start % chunk_depth == 0:
        lo = q_start
    else:
        lo = q_start - (q_start % chunk_depth)

    if q_stop % chunk_depth == 0:
        hi = q_stop
    else:
        hi = q_stop + (chunk_depth - q_stop % chunk_depth)

    return [lo + q_index, hi + q_index + 1]


def block_compute(x_start, x_stop,
                  y_start, y_stop,
                  z_start, z_stop,
                  origin=(0, 0, 1),
                  block_size=(256, 256, 16)):
    """
    Get bounding box coordinates (in 3D) of small cutouts to request in
    order to reconstitute a larger cutout.

    Arguments:
        :Q_start:               ``int`` The lower bound of dimension 'Q'
        :Q_stop:                ``int`` The upper bound of dimension 'Q'

    Returns:
        :[((x_start, x_stop), (y_start, y_stop), (z_start, z_stop)), ... ]:
    """
    # First, snap to block in each direction. Then we can divide sans
    # modulo without having to worry about remainders.
    bounds = [snap_to_cube(x_start, x_stop, block_size[0], origin[0]),
              snap_to_cube(y_start, y_stop, block_size[1], origin[1]),
              snap_to_cube(z_start, z_stop, block_size[2], origin[2])]

    chunks = []
    for q in range(3):
        chunks.append([(q_start, q_start + block_size[q]) for q_start in range(bounds[q][0], bounds[q][1], block_size[q])[:-1]])

    # There's a way to list-comprehensionify the above and the below, but...
    # a story for a different day.
    return [(chunks[0][x], chunks[1][y], chunks[2][z]) \
        for x in range(len(chunks[0])) \
        for y in range(len(chunks[1])) \
        for z in range(len(chunks[2]))]



from enums import *
from exceptions import *
import numpy

from RAMONBase import RAMONBase

class RAMONVolume(RAMONBase):
    """
    RAMONVolume Object for storing neuroscience data with a voxel volume
    """
    def __init__(self,
                xyz_offset = (0, 0, 0),
                resolution = 0,
                cutout = None,
                voxels = None,
                    id = DEFAULT_ID,
                    confidence = DEFAULT_CONFIDENCE,
                    dynamic_metadata = DEFAULT_DYNAMIC_METADATA,
                    status = DEFAULT_STATUS,
                    author = DEFAULT_AUTHOR):
        """
        Initialize a new RAMONVolume object. Inherit default attributes from RAMONBase,
        as well as:

        Arguments:
            xyz_offset      ``int[3] : (0, 0, 0)`` x,y,z coordinates of the minimum
                            corner of the cube (if data is a cutout), otherwise empty
            resolution      ``int : 0`` level in the database resolution hierarchy
            cutout          ``int`` dense matrix of data; left empty when using voxel lists
            voxels          ``int[][3]`` list of voxels making up annotation
        """

        self.xyz_offset = xyz_offset
        self.resolution = resolution
        self.cutout = cutout
        self.voxels = voxels

        RAMONBase.__init__(self, id=id, confidence=confidence,
                            dynamic_metadata=dynamic_metadata,
                            status=status, author=author)




class InvalidEnumerationException(Exception):
    def __init__(self, msg = "Invalid Enumeration."):
        self.msg = msg

    def __str__(self):
        return self.msg

def enum(*sequential, **named):
    enums = dict(zip(sequential, range(len(sequential))), **named)
    reverse = dict((value, key) for key, value in enums.iteritems())
    enums['reverse_mapping'] = reverse
    return type('Enum', (), enums)

eRAMONAnnoStatus = enum("DEFAULT",
                        UNPROCESSED=0,
                        LOCKED=1,
                        PROCESSED=2,
                        IGNORED=3)

DEFAULT_ID                  = -1
DEFAULT_CONFIDENCE          = 00
DEFAULT_DYNAMIC_METADATA    = {}
DEFAULT_STATUS              = eRAMONAnnoStatus.DEFAULT
DEFAULT_AUTHOR              = ''




class RemoteError(Exception):
    """
    A generic error arising from an issue pertaining
    to ndio Remotes.
    """
    def __init__(self, message):
        super(Exception, self).__init__(message)


class RemoteDataNotFoundError(RemoteError):
    """
    Called when data is requested from a Remote but the
    server either cannot access it (maybe a permissions issue?)
    or the data does not exist.
    """
    def __init__(self, message):
        super(RemoteError, self).__init__(message)


class RemoteDataUploadError(RemoteError):
    """
    Called when there's an issue during data upload. Could be
    because you don't have write access to this channel, or
    because there's a problem with your data. (Check your bounds?)
    """
    def __init__(self, message):
        super(RemoteError, self).__init__(message)


class BadAPIKeyError(RemoteError):
    """
    Called when you do not specify an API key and one is needed,
    or if you specify an API key that does not match the requirements
    to access/edit the data you're accessing/editing.
    """
    def __init__(self, message):
        super(RemoteError, self).__init__(message)



from enums import *
from exceptions import *
import numpy

from RAMONVolume import RAMONVolume

class RAMONGeneric(RAMONVolume):
    """
    RAMONGeneric Object for storing neuroscience data with a voxel volume
    """
    def __init__(self,
                xyz_offset = (0, 0, 0),
                resolution = 0,
                cutout = None,
                voxels = None,
                    id = DEFAULT_ID,
                    confidence = DEFAULT_CONFIDENCE,
                    dynamic_metadata = DEFAULT_DYNAMIC_METADATA,
                    status = DEFAULT_STATUS,
                    author = DEFAULT_AUTHOR):

            RAMONVolume.__init__(self,
                            xyz_offset = xyz_offset,
                            resolution = resolution,
                            cutout = cutout,
                            voxels = voxels,
                            id = id,
                            confidence = confidence,
                            dynamic_metadata = dynamic_metadata,
                            status = status,
                            author = author)



from enums import *
from exceptions import *

class RAMONBase(object):
    """
    RAMONBase Object for storing neuroscience data
    """
    def __init__(self,  id = DEFAULT_ID,
                        confidence = DEFAULT_CONFIDENCE,
                        dynamic_metadata = DEFAULT_DYNAMIC_METADATA,
                        status = DEFAULT_STATUS,
                        author = DEFAULT_AUTHOR):
        """
        Initialize a new RAMONBase object with default attributes.

        Arguments:
            :id:                `int` Unique 32bit ID value assigned by OCP database
            :confidence:        `float` Value 0-1 indicating confidence in annotation
            :dynamic_metadata:  `dict` A flexible, unspecified collection key-value pairs
            :status:            `string` Status of annotation in database
            :author:            `string` Username of the person who created the annotation
        """
        self._id = id
        self.confidence = confidence
        self.dynamic_metadata = dynamic_metadata
        self._status = status
        self.author = author


    @property
    def id(self):
        return self._id
    @id.setter
    def id(self, value):
        if type(value) is not int or value <= 0:
            raise InvalidIDException("Invalid ID {0}".format(value))
        self._id = value
        return value


    @property
    def status(self):
        return self._status
    @status.setter
    def status(self, value):
        try:
            r = eRAMONAnnoStatus.reverse_mapping[value]
        except:
            raise InvalidEnumerationException()
            return False
        self._status = value



from ndio.ramon.RAMONBase import *
from ndio.ramon.RAMONGeneric import *
from ndio.ramon.RAMONOrganelle import *
from ndio.ramon.RAMONSegment import *
from ndio.ramon.RAMONSynapse import *
from ndio.ramon.RAMONVolume import *


def metadata_to_ramon(metadata):
    """
    Get as much info about a RAMON object as possible
    from metadata.

    Might have been downloaded using a function like
    `ndio.remote.OCP.get_ramon_metadata()`, etc
    """
    raise NotImplementedError()
    pass




class DataSizeError(Exception):
    """
    An exception to raise when a dataset mismatches its Request,
    or more data is posted than there is 'space' for in the Request.
    Often a result of requesting data and failing with a broken pipe,
    or trying to POST 2D data into a 3D Request.
    """
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class Request(object):
    """
    A OCPRequest holds an absolute URL that can be used
    to access a unique datum. Requests can be thought of as
    a Python class wrapper for a RESTful endpoint.
    """
    def __init__(self, *args, **kwargs):
        """
        There are three ways of creating a Request; by URL,
        by filename, by an ordered tuple (in the order of the
        URL, incidentally) or by named variables. For example:

        ``Request('http://openconnecto.me/ocp/ca/tokenName/channelName/hdf5/1/0,10/0,10/0,10/')``

        ``Request('openconnecto.me-ocp-ca-tokenName-channelName-hdf5-1-0,10-0,10-0,10.hdf5')``

        ``Request('http://openconnecto.me', 'tokenName', 'channelName', 'hdf5', '1', '0', '10', '0', '10', '0', '10')``


        ``Request('server=http://openconnecto.me', token='tokenName', channel='channelName', format='hdf5', resolution='1', x_start='0', x_stop='10', y_start='0', y_stop='10', z_start='0', z_stop='10')``

        ...are all equivalent ways of creating a Request.

        These requests are currently all Image/Dense Requests... In a subsequent release, DenseRequest will inherit from a general Request class.
        """
        if len(args) == 1:
            if type(args[0]) is str:
                # Either URL or filename.
                url = args[0]
                if url.startswith('http'):
                    self._init_from_url(url)
                else:
                    self._init_from_filename(url)
            elif type(args[0]) is tuple:
                # Parse from an ordered tuple
                self._init_from_ordered_tuple(args[0])
            else:
                # There is no other way to create a Request from a single argument. Fail here.
                raise ValueError("Unsupported constructor for type {0}".format(type(args[0])))
        else:
            # We should expect named variables for each attribute.
            self._init_from_individual_values(**kwargs)

    def _init_from_filename(self, fname):
        # Remove file extension
        fname = ''.join(fname.split('.')[:-1])
        self._init_from_url(fname, delimiter="-")

    def _init_from_url(self, url, delimiter='/'):
        """
        protocol://server/ocp/ca/token/channel/fmt/res/x_start,x_stop/y_start,y_stop/z_start,z_stop/
        """
        # The delimiter argument is so that we can use the same
        # function for URL and Filename parsing.
        url = url.strip(delimiter)
        split_url = url.split('://')
        if len(split_url) == 2:
            self.protocol, url = split_url
        else:
            self.protocol = DEFAULT_PROTOCOL

        self.server, _, _, self.token, self.channel, self.format, self.resolution, x_range, y_range, z_range = url.split(delimiter)

        self.x_start, self.x_stop = x_range.split(',')
        self.y_start, self.y_stop = y_range.split(',')
        self.z_start, self.z_stop = z_range.split(',')

    def _init_from_ordered_tuple(self, tuple):
        raise NotImplementedError("Not yet implemented.")

    def _init_from_individual_values(self,
                        server="", token="", channel="",
                        x_start="", x_stop="",
                        y_start="", y_stop="", z_start="", z_stop="",
                        resolution="", format='hdf5'):
        split_server = server.split('://')
        if len(split_server) == 2:
            self.protocol, self.server = split_server
        elif len(split_server) == 1:
            self.protocol = DEFAULT_PROTOCOL
            self.server = server
        else:
            raise ValueError("Invalid server name: {0}".format(server))

        self.token = token
        self.channel = channel
        self.x_start = x_start
        self.x_stop = x_stop
        self.y_start = y_start
        self.y_stop = y_stop
        self.z_start = z_start
        self.z_stop = z_stop
        self.resolution = resolution
        self.format = format

    def to_url(self):
        """
        Output a URL that can be used to execute this Request.

        Arguments:
            None.
        Returns:
            ``str`` RESTful endpoint URL.
        """
        return '/'.join([
            self.protocol + ":/",
            self.server,
            "ocp",
            "ca",
            self.token,
            self.channel,
            self.format,
            str(self.resolution),
            str(self.x_start) + "," + str(self.x_stop),
            str(self.y_start) + "," + str(self.y_stop),
            str(self.z_start) + "," + str(self.z_stop),
        ""])


    def to_filename(self):
        """
        Output a filename that can be used to save data from this
        request and can later be used to create a new Request (using
        the _init_from_filename constructor) to download identical
        data. In that regard, we can use a filename as a unique
        'hash' of its data.

        Arguments:
            None.
        Returns:
            ``str`` Suggested filename for data that has been
            downloaded by this Request.
        """
        return '-'.join([
            self.server,
            "ocp",
            "ca",
            self.token,
            self.channel,
            self.format,
            str(self.resolution),
            str(self.x_start).zfill(6) + "," + str(self.x_stop).zfill(6),
            str(self.y_start).zfill(6) + "," + str(self.y_stop).zfill(6),
            str(self.z_start).zfill(6) + "," + str(self.z_stop).zfill(6) + "." + self.format
        ])



import requests
import h5py
import os
import numpy
from cStringIO import StringIO
import zlib
import tempfile

from Remote import Remote
from errors import *
import ndio.ramon as ramon

DEFAULT_HOSTNAME = "openconnecto.me"
DEFAULT_PROTOCOL = "http"


class OCP(Remote):

    def __init__(self, hostname=DEFAULT_HOSTNAME, protocol=DEFAULT_PROTOCOL):
        super(OCP, self).__init__(hostname, protocol)


    def ping(self):
        return super(OCP, self).ping('public_tokens/')

    def url(self):
        return super(OCP, self).url('/ocp/ca/')


    def get_public_tokens(self):
        """
        Get a list of public tokens available on this server.

        Arguments:
            None
        Returns:
            `str[]` list of public tokens
        """
        r = requests.get(self.url() + "public_tokens/")
        return r.json()


    def get_proj_info(self, token):
        """
        Returns the project info for a given token.

        Arguments:
            `token`:        `str` Token to return information for
        Returns:
            JSON representation of proj_info
        """
        r = requests.get(self.url() + "{}/info/".format(token))
        return r.json()


    def get_image(self, token, channel,
                  x_start, x_stop,
                  y_start, y_stop,
                  z_index, resolution=0):
        """
        Return a binary-encoded, decompressed image. You should probably
        check to see that your token has an `image` channel before running this,
        or it won't do anything interesting.

        Arguments:
            :token:                 ``string`` Token to identify data to download
            :channel:               ``string`` Channel to identify data to download
            :Q_start:               ``int`` The lower bound of dimension 'Q'
            :Q_stop:                ``int`` The upper bound of dimension 'Q'
            :z_index:               ``int`` The layer of data to return (1D)
            :resolution:            ``int : 1`` Resolution level
        Returns:
            :``str`` binary image data:
        """
        r = requests.get(self.url() +
            "{}/image/xy/{}/{},{}/{},{}/{}/".format(token, resolution,
            x_start, x_stop,
            y_start, y_stop,
            z_index))
        if r.status_code == 200:
            return r.content
        else:
            return r.status_code


    def get_cutout(self, token, channel,
                 x_start, x_stop,
                 y_start, y_stop,
                 z_start, z_stop,
                 resolution=1,
                 block_size=(256, 256, 16),
                 crop=False):
        """
        Get data from the OCP server.

        Arguments:
            :token:                 ``string`` Token to identify data to download
            :channel:               ``string`` Channel
            :resolution:            ``int : 1`` Resolution level
            :Q_start:               ``int`` The lower bound of dimension 'Q'
            :Q_stop:                ``int`` The upper bound of dimension 'Q'
            :block_size:            ``int(3)`` Block size of this dataset
            :crop:                  ``boolean : True`` whether or not to crop
                                    the volume before returning it

        Returns:
            :``numpy.ndarray``: Downloaded data.
        """

        # Get an array-of-tuples of blocks to request.
        from ndio.utils.parallel import block_compute, snap_to_cube
        small_chunks = block_compute(x_start, x_stop,
                                     y_start, y_stop,
                                     z_start, z_stop)

        # Each time we download a chunk, we'll store it in
        # this, in the format (block_origin, data)
        downloaded_chunks = []
        for c in small_chunks:
            downloaded_chunks.append((
                c, self._get_cutout_no_chunking(token, channel, resolution,
                        c[0][0], c[0][1], c[1][0], c[1][1], c[2][0], c[2][1])))

        x_bounds = snap_to_cube(x_start, x_stop, chunk_depth=256, q_index=0)
        y_bounds = snap_to_cube(y_start, y_stop, chunk_depth=256, q_index=0)
        z_bounds = snap_to_cube(z_start, z_stop, chunk_depth=16,  q_index=1)

        volume = numpy.zeros((
                x_bounds[1]-x_bounds[0],
                y_bounds[1]-y_bounds[0],
                z_bounds[1]-z_bounds[0]))


        # TODO: Optimize.
        for chunk in downloaded_chunks:
            x_range, y_range, z_range = chunk[0]
            xi = 0
            for x in range(x_range[0], x_range[1]):
                yi = 0
                for y in range(y_range[0], y_range[1]):
                    zi = 0
                    for z in range(z_range[0], z_range[1]):
                        volume[x-x_range[0]][y-y_range[0]][z-z_range[0]] = chunk[1][zi][xi][yi]
                        zi += 1
                    yi += 1
                xi += 1

        if crop == False:
            return volume
        else:
            raise NotImplementedError("Can't handle crops yet, sorry! :(")
            # we have to go get the bounds, subtract what they asked for,
            # and then return a sub-volume.
            x_start_trim = x_start - x_bounds[0]
            y_start_trim = y_start - y_bounds[0]
            z_start_trim = z_start - z_bounds[0]



    def _get_cutout_no_chunking(self, token, channel, resolution,
                                x_start, x_stop, y_start, y_stop,
                                z_start, z_stop):
        req = requests.get(self.url() +
                "{}/{}/hdf5/{}/{},{}/{},{}/{},{}/".format(
                    token, channel, resolution,
                    x_start, x_stop,
                    y_start, y_stop,
                    z_start, z_stop
                ))
        if req.status_code is not 200:
            raise IOError("Bad server response: {}".format(req.status_code))

        with tempfile.NamedTemporaryFile() as tmpfile:
            tmpfile.write(req.content)
            tmpfile.seek(0)
            h5file = h5py.File(tmpfile.name, "r")
            return h5file.get(channel).get('CUTOUT')[:]
        raise IOError("Failed to make tempfile.")



    def post_cutout(self, token, channel,
                    x_start, x_stop,
                    y_start, y_stop,
                    z_start, z_stop,
                    data,
                    dtype='',
                    resolution=0,
                    roll_axis=True):
        """
        Post a cutout to the server.

        Arguemnts:
            token
            channel
            q_start
            q_stop
            data:           A numpy array of data. Pass in (x, y, z)
            resolution:     Default resolution of the data
            roll_axis:      Default True. Pass False if you're supplying data in
                            (z, x, y) order.
            dtype:          Pass in datatype if you know it. Otherwise we'll
                            check the projinfo.
        Returns:
            True on success

        Throws:
            RemoteDataUploadError if there's an issue during upload.
        """

        datatype = self.get_proj_info(token)['channels'][channel]['datatype']
        if data.dtype.name != datatype:
            data = data.astype(datatype)

        if roll_axis:
            # put the z-axis first
            data = numpy.rollaxis(data, 2)

        data = numpy.expand_dims(data, axis=0)
        tempfile = StringIO()
        numpy.save(tempfile, data)

        compressed = zlib.compress(tempfile.getvalue())

        url = self.url() + "{}/{}/npz/{}/{},{}/{},{}/{},{}/".format(
            token, channel,
            resolution,
            x_start, x_stop,
            y_start, y_stop,
            z_start, z_stop
        )

        req = requests.post(url, data=compressed, headers={'Content-Type': 'application/octet-stream'})

        if req.status_code is not 200:
            raise RemoteDataUploadError(req.text)
        else:
            return True


    def get_ramon_ids(self, token, channel='annotation'):
        """
        Return a list of all IDs available for download from this token and
        channel.

        Arguments:
            token:          Project to use
            channel:        Channel to use (default 'annotation')
        Returns:
            int[]
        Throws:
            RemoteDataNotFoundError
        """

        req = requests.get(self.url() + "{}/{}/query/".format(token, channel))

        if req.status_code is not 200:
            raise RemoteDataNotFoundError('No query results for token {}.'.format(token))
        else:
            with tempfile.NamedTemporaryFile() as tmpfile:
                tmpfile.write(req.content)
                tmpfile.seek(0)
                h5file = h5py.File(tmpfile.name, "r")
                import pdb; pdb.set_trace()
                print "next..."



    def get_ramon(self, token, channel, anno_id, resolution, metadata_only=False):
        """
        Download a RAMON object by ID.

        Arguments:
            token:          Project to use
            channel:        The channel to use
            anno_id:        The ID of a RAMON object to gather
            resolution:     Resolution (if not working, try a higher num)
            metadata_only:  Defers to `get_ramon_metadata` instead
        Returns:
            ndio.ramon.RAMON
        Throws:
            RemoteDataNotFoundError
        """

        metadata = self.get_ramon_metadata(token, channel, anno_id)

        if metadata_only:
            return metadata

        metadata = metadata[str(anno_id)]


        # Download the data itself


        req = requests.get(self.url() +
                "{}/{}/{}/cutout/{}".format(token, channel,
                anno_id, resolution))

        if req.status_code is not 200:
            raise RemoteDataNotFoundError('No data for id {}.'.format(anno_id))
        else:
            # Populate metadata of the RAMON Object
            if metadata['type'] == 'segment':
                # RAMONSegment
                r = ramon.RAMONSegment()
                r.segment_class =   metadata['metadata']['segmentclass']
                r.status =          metadata['metadata']['status']
                r.confidence =      metadata['metadata']['confidence']
                r.author =          metadata['metadata']['author']
                r.neuron =          metadata['metadata']['neuron']

                with tempfile.NamedTemporaryFile() as tmpfile:
                    tmpfile.write(req.content)
                    tmpfile.seek(0)
                    h5file = h5py.File(tmpfile.name, "r")

                    metadata = h5file[str(anno_id)]['METADATA']
                    r.author =          metadata['AUTHOR'][0]
                    r.confidence =      metadata['CONFIDENCE'][0]
                    r.neuron =          metadata['NEURON'][0]
                    r.parent_seed =     metadata['PARENTSEED'][0]
                    r.segment_class =   metadata['SEGMENTCLASS'][0]
                    r.status =          metadata['STATUS'][0]
                    r.synapses =        metadata['SYNAPSES'][()]
                    r.xyz_offset =      h5file[str(anno_id)]['XYZOFFSET'][()]
                    r.resolution =      h5file[str(anno_id)]['RESOLUTION'][0]
                    r.cutout =          h5file[str(anno_id)]['CUTOUT'][()]

                return r

            else:
                raise NotImplementedError("Only RAMONSegments are currently supported.")



    def get_ramon_metadata(self, token, channel, anno_id):
        """
        Download a RAMON object by ID. `anno_id` can be a string `"123"`, an
        int `123`, an array of ints `[123, 234, 345]`, an array of strings
        `["123", "234", "345"]`, or a comma-separated string list `"123,234,345"`.

        Arguments:
            token:      Project to use
            channel:    The channel to use
            anno_id:    An int, a str, or a list of ids to gather

        Returns:
            JSON. If you pass a single id in str or int, returns a single datum.
            If you pass a list of int or str or a comma-separated string, will
            return a dict with keys from the list and the values are the server-
            returned JSON.

        Throws:
            RemoteDataNotFoundError
        """

        if type(anno_id) is int:
            # there's just one ID to download
            return self._get_single_ramon_metadata(token, channel, str(anno_id))
        elif type(anno_id) is str:
            # either "id" or "id,id,id":
            if (len(anno_id.split(',')) > 1):
                results = {}
                for i in anno_id.split(','):
                    results[i] = self._get_single_ramon_metadata(token, channel,
                                                            anno_id.strip())
                return results
            else:
                # "id"
                return self._get_single_ramon_metadata(token, channel, anno_id.strip())
        elif type(anno_id) is list:
            # [id, id] or ['id', 'id']
            results = {}
            for i in anno_id:
                results[i] = self._get_single_ramon_metadata(token, channel,
                                                        str(anno_id).strip())
            return results



    def _get_single_ramon_metadata(self, token, channel, anno_id):
        req = requests.get(self.url() +
                "{}/{}/{}/json/".format(token, channel,
                anno_id))

        if req.status_code is not 200:
            raise RemoteDataNotFoundError('No data for id {}.'.format(anno_id))
        else:
            return req.json()


    def post_ramon(self, token, channel, ramon, opts=""):
        """
        Download a RAMON object by ID.

        Arguments:
            token:      Project to use
            channel:    The channel to use
            ramon:      A ndio.ramon.RAMON object
            opts:       String options (ignored)
        Returns:
            ndio.ramon.RAMON
        Throws:
            RemoteDataNotFoundError
        """

        req = requests.get(self.url() +
                "{}/{}/{}/".format(token, channel, opts))

        if req.status_code is not 200:
            raise RemoteDataNotFoundError('No data for id {}.'.format(anno_id))
        else:
            return True



import h5py
import numpy
import os

def import_hdf5(hdf5_filename):
    """
    Import a HDF5 file into a numpy array.

    Arguments:
        hdf5_filename:  A string filename of a HDF5 datafile

    Returns:
        A numpy array with data from the HDF5 file
    """

    # Expand filename to be absolute
    hdf5_filename = os.path.expanduser(hdf5_filename)

    try:
        f = h5py.File(hdf5_filename, "r")
        # OCP stores data inside the 'cutout' h5 dataset
        data_layers = f.get('image').get('CUTOUT')
    except Exception as e:
        print("Could not load file {0} for conversion.".format(hdf5_filename))
        raise

    return numpy.array(data_layers)



def export_hdf5(hdf5_filename, numpy_data):
    """
    Export a numpy array to a HDF5 file.

    Arguments:
        hdf5_filename:  A filename to which to save the HDF5 data
        numpy_data:     The numpy array to save to HDF5

    Returns:
        String. The expanded filename that now holds the HDF5 data
    """

    # Expand filename to be absolute
    hdf5_filename = os.path.expanduser(hdf5_filename)

    if os.path.exists(hdf5_filename):
        print("File {0} already exists, stopping...".format(hdf5_filename))
        return False

    try:
        h = h5py.File(hdf5_filename, "w")
        h.create_dataset('CUTOUT', data=numpy_data)
        h.close()
    except Exception as e:
        print("Could not save HDF5 file {0}.".format(hdf5_filename))

    return hdf5_filename



from PIL import Image
import numpy
import os
import glob

def import_png(png_filename):
    """
    Import a png file into a numpy array.

    Arguments:
        :png_filename:  A string filename of a png datafile

    Returns:
        A numpy array with data from the png file
    """

    # Expand filename to be absolute
    png_filename = os.path.expanduser(png_filename)

    try:
        img = Image.open(png_filename)
    except Exception as e:
        print("Could not load file {0} for conversion.".format(png_filename))
        raise

    return numpy.array(img)



def export_png(png_filename, numpy_data):
    """
    Export a numpy array to a png file.

    Arguments:
        png_filename:   A filename to which to save the png data
        numpy_data:     The numpy array to save to png. OR a string

    Returns:
        String. The expanded filename that now holds the png data
    """

    # Expand filename to be absolute
    png_filename = os.path.expanduser(png_filename)


    if type(numpy_data) is str:
        fp = open(png_filename, "wb")
        fp.write(numpy_data)
        fp.close()
        return png_filename

    try:
        if numpy_data.dtype.name != 'uint8':
            m = 'I'
            img = Image.fromarray(numpy_data, mode=m)
        else:
            img = Image.fromarray(numpy_data)
        img.save(png_filename)
    except Exception as e:
        print("Could not save png file {0}.".format(png_filename))
    return png_filename



def export_png_collection(png_filename_base, numpy_data, start_layers_at=1):
    """
    Export a numpy array to a set of png files, with each Z-index 2D
    array as its own 2D file.

    Arguments:
        png_filename_base:     A filename template, such as "my-image-*.png"
                                which will lead to a collection of files named
                                "my-image-0.png", "my-image-1.png", etc.
        numpy_data:             The numpy array data to save to png.

    Returns:
        Array. A list of expanded filenames that hold png data.
    """

    file_extension = png_filename_base.split('.')[-1]
    if file_extension in ['png']:
        # Filename is "name*.tif[f]", set file_base to "name*".
        file_base = '.'.join(png_filename_base.split('.')[:-1])
    else:
        # Filename is "name*", set file_base to "name*".
        # That is, extension wasn't included.
        file_base = png_filename_base
        file_extension = ".png"

    file_base_array = file_base.split('*')

    # The array of filenames to return
    output_files = []

    # Filename 0-padding
    i = start_layers_at
    for layer in numpy_data:
        layer_filename = (str(i).zfill(6)).join(file_base_array) + file_extension
        output_files.append(export_png(layer_filename, layer))
        i += 1

    return output_files

def import_png_collection(png_filename_base):
    """
    Import all files matching the filename base given via `png_filename_base`.
    Images are ordered by alphabetical order, which means that you *MUST* 0-pad
    your numbers if they span a power of ten (e.g. 0999-1000 or 09-10). This is
    handled automatically by the complement function, `export_png_collection`.
    Also, look at how nicely these documentation lines are all the same length!

    Arguments:
        png_filename_base:     An asterisk-wildcard string that should refer
                                to all TIFFs in the stack. All * are replaced
                                according to command-line expansion rules.
    Returns:
        A numpy array holding a 3D dataset
    """

    # We expect images to be indexed by their alphabetical order.
    files = glob.glob(png_filename_base)
    files.sort()

    numpy_data = []
    for f in files:
        numpy_data.append(import_png(f))

    return numpy_data
